########
#
# read_exif.py
#
# Given a folder of images, read relevant metadata (EXIF/IPTC/XMP) fields from all images, 
# and write them to  a .json or .csv file.  
#
# This module can use either PIL (which can only reliably read EXIF data) or exiftool (which
# can read everything).  The latter approach expects that exiftool is available on the system
# path.  No attempt is made to be consistent in format across the two approaches.
#
########

#%% Imports and constants

import os
import subprocess
import json
from datetime import datetime

from multiprocessing.pool import ThreadPool as ThreadPool
from multiprocessing.pool import Pool as Pool

from tqdm import tqdm
from PIL import Image, ExifTags

from md_utils.path_utils import find_images
from md_utils.ct_utils import args_to_object

debug_max_images = None


#%% Options

class ReadExifOptions:
    
    verbose = False
    
    # If this is True and an output file is specified for read_exif_from_folder,
    # and we encounter a serialization issue, we'll return the results but won't
    # error.    
    allow_write_error = False
    
    # Number of concurrent workers
    n_workers = 1
    
    # Should we use threads (vs. processes) for parallelization?
    #
    # Not relevant if n_workers is 1.
    use_threads = True
    
    tag_types_to_ignore = set(['File','ExifTool'])
    
    exiftool_command_name = 'exiftool'
    
    # How should we handle byte-formatted EXIF tags?
    #
    # 'convert_to_string': convert to a Python string
    # 'delete': don't include at all
    # 'raw': include as a byte string
    byte_handling = 'convert_to_string' # 'convert_to_string','delete','raw'
    
    # Should we use exiftool or pil?
    processing_library = 'pil' # 'exiftool','pil'


#%% Functions

def enumerate_files(input_folder):
    """
    Enumerates all image files in input_folder, returning relative paths
    """
    
    image_files = find_images(input_folder,recursive=True)
    image_files = [os.path.relpath(s,input_folder) for s in image_files]
    image_files = [s.replace('\\','/') for s in image_files]
    print('Enumerated {} files'.format(len(image_files)))
    return image_files


def get_exif_ifd(exif):
    """
    Read EXIF data by finding the EXIF offset and reading tags directly
    
    https://github.com/python-pillow/Pillow/issues/5863
    """
    
    # Find the offset for all the EXIF information
    for key, value in ExifTags.TAGS.items():
        if value == "ExifOffset":
            break
    info = exif.get_ifd(key)
    return {
        ExifTags.TAGS.get(key, key): value
        for key, value in info.items()
    }


def read_pil_exif(im,options=None):
    """
    Read all the EXIF data we know how to read from [im] (path or PIL Image), whether it's 
    in the PIL default EXIF data or not.
    """
    
    if options is None:
        options = ReadExifOptions()
        
    image_name = '[image]'
    if isinstance(im,str):
        image_name = im
        im = Image.open(im)
        
    exif_tags = {}
    try:
        exif_info = im.getexif()
    except Exception:
        exif_info = None
        
    if exif_info is None:
        return exif_tags
    
    for k, v in exif_info.items():
        assert isinstance(k,str) or isinstance(k,int), \
            'Invalid EXIF key {}'.format(str(k))
        if k in ExifTags.TAGS:
            exif_tags[ExifTags.TAGS[k]] = str(v)
        else:
            # print('Warning: unrecognized EXIF tag: {}'.format(k))
            exif_tags[k] = str(v)
    
    exif_idf_tags = get_exif_ifd(exif_info)
    
    for k in exif_idf_tags.keys():
        v = exif_idf_tags[k]
        if k in exif_tags:
            if options.verbose:
                print('Warning: redundant EXIF values for {} in {}:\n{}\n{}'.format(
                    k,image_name,exif_tags[k],v))
        else:
            exif_tags[k] = v
    
    exif_tag_names = list(exif_tags.keys())
    
    # Type conversion and cleanup
    # 
    # Most quirky types will get serialized to string when we write to .json.
    for k in exif_tag_names:
        
        if isinstance(exif_tags[k],bytes):
            
            if options.byte_handling == 'delete':
                del exif_tags[k]
            elif options.byte_handling == 'raw':
                pass
            else:
                assert options.byte_handling == 'convert_to_string'
                exif_tags[k] = str(exif_tags[k])
        
        elif isinstance(exif_tags[k],str):
            
            exif_tags[k] = exif_tags[k].strip()
            
    return exif_tags

# ...read_pil_exif()


def format_datetime_as_exif_datetime_string(dt):
    """
    Returns a Python datetime object rendered using the standard Exif datetime
    string format
    """
    
    return datetime.strftime(dt, '%Y:%m:%d %H:%M:%S')
    

def parse_exif_datetime_string(s,verbose=False):
    """"
    Exif datetimes are strings, but in a standard format:        
        
    %Y:%m:%d %H:%M:%S
    
    Parse one of those strings into a Python datetime object.
    """
    
    dt = None
    try:
        dt = datetime.strptime(s, '%Y:%m:%d %H:%M:%S')
    except Exception:
        if verbose:
            print('Warning: could not parse datetime {}'.format(str(s)))
    return dt


def read_exif_tags_for_image(file_path,options=None):
    """
    Get relevant fields from EXIF data for an image
    
    Returns a dict with fields 'status' (str) and 'tags'
    
    The exact format of 'tags' depends on options.processing_library
    
    For exiftool, 'tags' is a list of lists, where each element is (type/tag/value)
    
    For pil, 'tags' is a dict (str:str)
    """
    
    if options is None:
        options = ReadExifOptions()
    
    result = {'status':'unknown','tags':[]}
    
    if options.processing_library == 'pil':
        
        try:
            exif_tags = read_pil_exif(file_path,options)

        except Exception as e:
            if options.verbose:
                print('Read failure for image {}: {}'.format(
                    file_path,str(e)))
            result['status'] = 'read_failure'
            result['error'] = str(e)
        
        if result['status'] == 'unknown':
            if exif_tags is None:            
                result['status'] = 'empty_read'
            else:
                result['status'] = 'success'
                result['tags'] = exif_tags
                
        return result
        
    elif options.processing_library == 'exiftool':
        
        # -G means "Print group name for each tag", e.g. print:
        #
        # [File]          Bits Per Sample                 : 8
        #
        # ...instead of:
        #
        # Bits Per Sample                 : 8
        proc = subprocess.Popen([options.exiftool_command_name, '-G', file_path],
                                stdout=subprocess.PIPE, encoding='utf8')
        
        exif_lines = proc.stdout.readlines()    
        exif_lines = [s.strip() for s in exif_lines]
        if ( (exif_lines is None) or (len(exif_lines) == 0) or not \
            any([s.lower().startswith('[exif]') for s in exif_lines])):
            result['status'] = 'failure'
            return result
        
        # A list of three-element lists (type/tag/value)
        exif_tags = []
        
        # line_raw = exif_lines[0]
        for line_raw in exif_lines:
            
            # A typical line:
            #
            # [ExifTool]      ExifTool Version Number         : 12.13
            
            line = line_raw.strip()
            
            # Split on the first occurrence of ":"
            tokens = line.split(':',1)
            assert(len(tokens) == 2), 'EXIF tokenization failure ({} tokens, expected 2)'.format(
                len(tokens))
            
            field_value = tokens[1].strip()        
            
            field_name_type = tokens[0].strip()        
            field_name_type_tokens = field_name_type.split(None,1)
            assert len(field_name_type_tokens) == 2, 'EXIF tokenization failure'
            
            field_type = field_name_type_tokens[0].strip()
            assert field_type.startswith('[') and field_type.endswith(']'), \
                'Invalid EXIF field {}'.format(field_type)
            field_type = field_type[1:-1]
            
            if field_type in options.tag_types_to_ignore:
                if options.verbose:
                    print('Ignoring tag with type {}'.format(field_type))
                continue        
            
            field_tag = field_name_type_tokens[1].strip()
            
            tag = [field_type,field_tag,field_value]
            
            exif_tags.append(tag)
            
        # ...for each output line
            
        result['status'] = 'success'
        result['tags'] = exif_tags
        return result
    
    else:
        
        raise ValueError('Unknown processing library {}'.format(
            options.processing_library))

    # ...which processing library are we using?
    
# ...read_exif_tags_for_image()


def populate_exif_data(im, image_base, options=None):
    """
    Populate EXIF data into the 'exif_tags' field in the image object [im].
    
    im['file_name'] should be prepopulated, relative to image_base.
    
    Returns a modified version of [im], also modifies [im] in place.
    """
    
    if options is None:
        options = ReadExifOptions()

    fn = im['file_name']
    if options.verbose:
        print('Processing {}'.format(fn))
    
    try:
        
        file_path = os.path.join(image_base,fn)
        assert os.path.isfile(file_path), 'Could not find file {}'.format(file_path)
        result = read_exif_tags_for_image(file_path,options)
        if result['status'] == 'success':
            exif_tags = result['tags']            
            im['exif_tags'] = exif_tags
        else:
            im['exif_tags'] = None
            im['status'] = result['status']
            if 'error' in result:
                im['error'] = result['error']
            if options.verbose:
                print('Error reading EXIF data for {}'.format(file_path))
    
    except Exception as e:
        
        s = 'Error on {}: {}'.format(fn,str(e))
        print(s)
        im['error'] = s
        im['status'] = 'read failure'
        im['exif_tags'] = None
    
    return im

# ...populate_exif_data()


def create_image_objects(image_files):
    """
    Create empty image objects for every image in [image_files], which can be a 
    list of relative paths (which will get stored without processing, so the base 
    path doesn't matter here), or a folder name.
    
    Returns a list of dicts with field 'file_name' (a relative path).
    """
    
    # Enumerate *relative* paths
    if isinstance(image_files,str):    
        print('Enumerating image files in {}'.format(image_files))
        assert os.path.isdir(image_files), 'Invalid image folder {}'.format(image_files)
        image_files = enumerate_files(image_files)
        
    images = []
    for fn in image_files:
        im = {}
        im['file_name'] = fn
        images.append(im)
    
    if debug_max_images is not None:
        print('Trimming input list to {} images'.format(debug_max_images))
        images = images[0:debug_max_images]
    
    return images


def populate_exif_for_images(image_base,images,options=None):
    """
    Main worker loop: read EXIF data for each image object in [images] and 
    populate the image objects.
    
    'images' should be a list of dicts with the field 'file_name' containing
    a relative path (relative to 'image_base').    
    """
    
    if options is None:
        options = ReadExifOptions()

    if options.n_workers == 1:
      
        results = []
        for im in tqdm(images):
            results.append(populate_exif_data(im,image_base,options))
        
    else:
        
        from functools import partial
        if options.use_threads:
            print('Starting parallel thread pool with {} workers'.format(options.n_workers))
            pool = ThreadPool(options.n_workers)
        else:
            print('Starting parallel process pool with {} workers'.format(options.n_workers))
            pool = Pool(options.n_workers)
    
        results = list(tqdm(pool.imap(partial(populate_exif_data,image_base=image_base,
                                        options=options),images),total=len(images)))

    return results


def write_exif_results(results,output_file):
    """
    Write EXIF information to [output_file].
    
    'results' is a list of dicts with fields 'exif_tags' and 'file_name'.

    Writes to .csv or .json depending on the extension of 'output_file'.         
    """
    
    if output_file.endswith('.json'):
        
        with open(output_file,'w') as f:
            json.dump(results,f,indent=1,default=str)
            
    elif output_file.endswith('.csv'):
        
        # Find all EXIF tags that exist in any image
        all_keys = set()
        for im in results:
            
            keys_this_image = set()
            exif_tags = im['exif_tags']
            file_name = im['file_name']
            for tag in exif_tags:
                tag_name = tag[1]
                assert tag_name not in keys_this_image, \
                    'Error: tag {} appears twice in image {}'.format(
                        tag_name,file_name)
                all_keys.add(tag_name)
                
            # ...for each tag in this image
            
        # ...for each image
        
        all_keys = sorted(list(all_keys))
        
        header = ['File Name']
        header.extend(all_keys)
        
        import csv
        with open(output_file,'w') as csvfile:
            
            writer = csv.writer(csvfile)
            
            # Write header
            writer.writerow(header)
            
            for im in results:
                
                row = [im['file_name']]
                kvp_this_image = {tag[1]:tag[2] for tag in im['exif_tags']}
                
                for i_key,key in enumerate(all_keys):
                    value = ''
                    if key in kvp_this_image:
                        value = kvp_this_image[key]
                    row.append(value)                                        
                # ...for each key that *might* be present in this image
                
                assert len(row) == len(header)
                
                writer.writerow(row)
                
            # ...for each image
            
        # ...with open()
    
    else:
        
        raise ValueError('Could not determine output type from file {}'.format(
            output_file))
        
    # ...if we're writing to .json/.csv
    
    print('Wrote results to {}'.format(output_file))


def is_executable(name):
    
    """Check whether `name` is on PATH and marked as executable."""
    
    # https://stackoverflow.com/questions/11210104/check-if-a-program-exists-from-a-python-script

    from shutil import which
    return which(name) is not None


def read_exif_from_folder(input_folder,output_file=None,options=None,filenames=None):
    """
    Read EXIF data for all images in input_folder.
    
    If filenames is not None, it should be a list of relative filenames; only those files will 
    be processed.
    
    input_folder can be None or '', in which case filenames should be a list of absolute paths.
    
    if output_file is not None, results will be written to the specified .json file.
    
    returns a dictionary mapping relative filenames to EXIF data.
    """
    
    if options is None:
        options = ReadExifOptions()
    
    if input_folder is None:
        input_folder = ''
    if len(input_folder) > 0:
        assert os.path.isdir(input_folder), \
            '{} is not a valid folder'.format(input_folder)

    assert (len(input_folder) > 0) or (filenames is not None), \
        'Must specify either a folder or a list of files'
        
    if output_file is not None:    
        
        assert output_file.lower().endswith('.json') or output_file.lower().endswith('.csv'), \
            'I only know how to write results to .json or .csv'
            
        try:
            with open(output_file, 'a') as f:
                if not f.writable():
                    raise IOError('File not writable')
        except Exception:
            print('Could not write to file {}'.format(output_file))
            raise
        
    if options.processing_library == 'exif':
        assert is_executable(options.exiftool_command_name), 'exiftool not available'

    if filenames is None:
        images = create_image_objects(input_folder)
    else:
        assert isinstance(filenames,list)
        images = create_image_objects(filenames)
        
    results = populate_exif_for_images(input_folder,images,options)
    
    if output_file is not None:
        try:
            write_exif_results(results,output_file)
        except Exception as e:
            if not options.allow_write_error:
                raise
            else:
                print('Warning: error serializing EXIF data: {}'.format(str(e)))                
        
    return results

    
#%% Interactive driver

if False:
    
    #%%
    
    input_folder = os.path.expanduser('~/data/KRU-test')
    output_file = os.path.expanduser('~/data/test-exif.json')
    options = ReadExifOptions()
    options.verbose = False
    options.n_workers = 10
    options.use_threads = False
    options.processing_library = 'pil'
    # options.processing_library = 'exiftool'
    
    results = read_exif_from_folder(input_folder,output_file,options)

    #%%
    
    with open(output_file,'r') as f:
        d = json.load(f)
        

#%% Command-line driver

import argparse
import sys

def main():

    options = ReadExifOptions()
    
    parser = argparse.ArgumentParser(description=('Read EXIF information from all images in' + \
                                                  ' a folder, and write the results to .csv or .json'))

    parser.add_argument('input_folder', type=str)
    parser.add_argument('output_file', type=str)
    parser.add_argument('--n_workers', type=int, default=1,
                        help='Number of concurrent workers to use (defaults to 1)')
    parser.add_argument('--use_threads', action='store_true',
                        help='Use threads (instead of processes) for multitasking')
    parser.add_argument('--processing_library', type=str, default=options.processing_library,
                        help='Processing library (exif or pil)')
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()    
    args_to_object(args, options)
    options.processing_library = options.processing_library.lower()
    
    read_exif_from_folder(args.input_folder,args.output_file,options)
    
if __name__ == '__main__':
    main()



########
#
# cct_to_csv.py
#
# "Converts" a COCO Camera Traps .json file to .csv, in quotes because 
# all kinds of assumptions are made here, and if you have a particular .csv
# format in mind, YMMV.  Most notably, does not include any bounding box information
# or any non-standard fields that may be present in the .json file.  Does not
# propagate information about sequence-level vs. image-level annotations.
#
# Does not assume access to the images, therefore does not open .jpg files to find
# datetime information if it's not in the metadata, just writes datetime as 'unknown'.
#
########

#%% Imports

import os
import json
from tqdm import tqdm
from collections import defaultdict 


#%% Main function

def cct_to_csv(input_file,output_file=None):

    if output_file is None:
        output_file = input_file + '.csv'

    ##%% Read input
    
    print('Loading input data')
    
    with open(input_file,'r') as f:
        input_data = json.load(f)
    
    
    ##%% Build internal mappings
    
    print('Processing input data')    
    
    images = input_data['images']
    
    category_id_to_name = {cat['id']:cat['name'] for cat in input_data['categories']}
    
    image_id_to_class_names = defaultdict(set)
    
    annotations = input_data['annotations']
                             
    # annotation = annotations[0]
    for annotation in tqdm(annotations):
        image_id = annotation['image_id']
        class_name = annotation['category_id']
        image_id_to_class_names[image_id].add(
            category_id_to_name[class_name])
        
    
    ##%% Write output file
    
    print('Writing output file')
    
    with open(output_file,'w') as f:
        
        f.write('relative_path,datetime,location,sequence_id,class_name\n')
        
        # im = images[0]
        for im in tqdm(images):
            
            file_name = im['file_name']
            class_names_set = image_id_to_class_names[im['id']]
            assert len(class_names_set) > 0
            
            if 'datetime' in im:
                datetime = im['datetime']
            else:
                datetime = 'unknown'
            
            if 'location' in im:
                location = im['location']
            else:
                location = 'unknown'
            
            if 'seq_id' in im:
                sequence_id = im['seq_id']
            else:
                sequence_id = 'unknown'
            
            # Write out one line per class:
            for class_name in class_names_set:
                f.write('{},{},{},{},{}\n'.format(file_name,
                   datetime,location,sequence_id,class_name))
            
            # ...for each class name
            
        # ...for each image
        
    # ...with open(output_file)    
            
# ...def cct_to_csv


#%% Interactive driver

if False:

    #%%
    
    input_dir = r"G:\temp\cct-to-csv"
    files = os.listdir(input_dir)
    files = [s for s in files if s.endswith('.json')]
    for fn in files:
        input_file = os.path.join(input_dir,fn)
        assert os.path.isfile(input_file)
        cct_to_csv(input_file)


#%% Command-line driver

import argparse

def main():

    parser = argparse.ArgumentParser(description=('"Convert" a COCO Camera Traps .json file to .csv (read code to see why "convert" is in quotes)'))

    parser.add_argument('input_file', type=str)                        
    parser.add_argument('--output_file', type=str, default=None)
    
    args = parser.parse_args()
    cct_to_csv(args.input_file,args.output_file)
    
if __name__ == '__main__':
    main()



########
#
# add_width_and_height_to_db.py
#
# Grabs width and height from actual image files for a .json database that is missing w/h.
#
########

#%% Imports and constants

import json
from PIL import Image

datafile = '/datadrive/snapshotserengeti/databases/snapshotserengeti.json'
image_base = '/datadrive/snapshotserengeti/images/'


#%% Execution

with open(datafile,'r') as f:
    data = json.load(f)

for im in data['images']:
    if 'height' not in im:
        im_w, im_h = Image.open(image_base+im['file_name']).size
        im['height'] = im_h
        im['width'] = im_w

json.dump(data, open(datafile,'w'))



########
#
# idfg_iwildcam_lila_prep.py
#
# Adding class labels (from the private test .csv) to the iWildCam 2019 IDFG 
# test set, in preparation for release on LILA.
#
# This version works with the public iWildCam release images.
#
########

#%% ############ Take one, from iWildCam .json files ############ 

#%% Imports and constants

import uuid
import json
import os
from tqdm import tqdm

base_folder = r'h:\iWildCam_2019_IDFG'
input_json = os.path.join(base_folder,'iWildCam_2019_IDFG_info.json')
input_csv = os.path.join(base_folder,'IDFG_eval_public_v_private.csv')
output_json = os.path.join(base_folder,'idaho_camera_traps.json')

assert os.path.isfile(input_json)
assert os.path.isfile(input_csv)


#%% Read input files

with open(input_json,'r') as f:
    input_data = json.load(f)

with open(input_csv,'r') as f:
    private_csv_lines = f.readlines()

private_csv_lines = [s.strip() for s in private_csv_lines]
    
# Remove the header line    
assert private_csv_lines[0] == 'Id,Category,Usage'
private_csv_lines = private_csv_lines[1:]

print('Read {} annotations for {} images'.format(len(private_csv_lines),len(input_data['images'])))

assert len(private_csv_lines) == len(input_data['images'])
n_images = len(input_data['images'])


#%% Parse annotations

image_id_to_category_ids = {}
for line in tqdm(private_csv_lines):
    
    # Lines look like:
    #
    # b005e5b2-2c0b-11e9-bcad-06f1011196c4,1,Private
    
    tokens = line.split(',')
    assert len(tokens) == 3
    assert tokens[2] in ['Private','Public']
    image_id_to_category_ids[tokens[0]] = int(tokens[1])
    
assert len(image_id_to_category_ids) == n_images


#%% Minor cleanup re: images

for im in tqdm(input_data['images']):
    image_id = im['id']
    im['file_name'] = im['file_name'].replace('iWildCam_IDFG_images/','')
    assert isinstance(im['location'],int)
    im['location'] = str(im['location'])
    

#%% Create annotations

annotations = []

for image_id in tqdm(image_id_to_category_ids):
    category_id = image_id_to_category_ids[image_id]
    ann = {}
    ann['id'] = str(uuid.uuid1())
    ann['image_id'] = image_id
    ann['category_id'] = category_id
    annotations.append(ann)
    

#%% Prepare info

info = input_data['info']
info['contributor'] = 'Images acquired by the Idaho Department of Fish and Game, dataset curated by Sara Beery'
info['description'] = 'Idaho Camera traps'
info['version'] = '2021.07.19'


#%% Minor adjustments to categories

input_categories = input_data['categories']

category_id_to_name = {cat['id']:cat['name'] for cat in input_categories}
category_name_to_id = {cat['name']:cat['id'] for cat in input_categories}
assert category_id_to_name[0] == 'empty'

category_names_to_counts = {}
for category in input_categories:
    category_names_to_counts[category['name']] = 0
    
for ann in annotations:
    category_id = ann['category_id']
    category_name = category_id_to_name[category_id]
    category_names_to_counts[category_name] = category_names_to_counts[category_name] + 1
    
categories = []

for category_name in category_names_to_counts:
    count = category_names_to_counts[category_name]    
    
    # Remove unused categories
    if count == 0:
        continue
    
    category_id = category_name_to_id[category_name]
    
    # Name adjustments
    if category_name == 'prongs':
        category_name = 'pronghorn'
    
    categories.append({'id':category_id,'name':category_name})
    
    
#%% Create output

output_data = {}
output_data['images'] = input_data['images']
output_data['annotations'] = annotations
output_data['categories'] = categories
output_data['info'] = info


#%% Write output

with open(output_json,'w') as f:
    json.dump(output_data,f,indent=2)
    

#%% Validate .json file

from data_management.databases import sanity_check_json_db

options = sanity_check_json_db.SanityCheckOptions()
options.baseDir = os.path.join(base_folder,'images'); assert os.path.isdir(options.baseDir)
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

_, _, _ = sanity_check_json_db.sanity_check_json_db(output_json, options)


#%% Preview labels

from md_visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 100
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

# viz_options.classes_to_exclude = ['test']
html_output_file, _ = visualize_db.process_images(db_path=output_json,
                                                         output_dir=os.path.join(
                                                         base_folder,'preview'),
                                                         image_base_dir=os.path.join(base_folder,'images'),
                                                         options=viz_options)
os.startfile(html_output_file)


#%% ############ Take two, from pre-iWildCam .json files created from IDFG .csv files ############ 

#%% Imports and constants

import json
import os

base_folder = r'h:\idaho-camera-traps'
input_json_sl = os.path.join(base_folder,'iWildCam_IDFG.json')
input_json = os.path.join(base_folder,'iWildCam_IDFG_ml.json')
output_json = os.path.join(base_folder,'idaho_camera_traps.json')
remote_image_base_dir = r'z:\idfg'

assert os.path.isfile(input_json)


#%% One-time line break addition

if not os.path.isfile(input_json):
    
    sl_json = input_json_sl
    ml_json = input_json
    
    with open(sl_json,'r') as f:
        d = json.load(f)
    with open(ml_json,'w') as f:
        json.dump(d,f,indent=2)
        
        
#%% Read input files

with open(input_json,'r') as f:
    input_data = json.load(f)

print('Read {} annotations for {} images'.format(len(input_data['annotations']),len(input_data['images'])))

n_images = len(input_data['images'])


#%% Prepare info

info = {}
info['contributor'] = 'Images acquired by the Idaho Department of Fish and Game, dataset curated by Sara Beery'
info['description'] = 'Idaho Camera traps'
info['version'] = '2021.07.19'


#%% Minor adjustments to categories

input_categories = input_data['categories']
output_categories = []

for c in input_categories:
    category_name = c['name']
    category_id = c['id']
    if category_name == 'prong':
        category_name = 'pronghorn'
    category_name = category_name.lower()
    output_categories.append({'name':category_name,'id':category_id})


#%% Minor adjustments to annotations

for ann in input_data['annotations']:
    ann['id'] = str(ann['id'])


#%% Create output

output_data = {}
output_data['images'] = input_data['images']
output_data['annotations'] = input_data['annotations']
output_data['categories'] = output_categories
output_data['info'] = info


#%% Write output

with open(output_json,'w') as f:
    json.dump(output_data,f,indent=2)
    

#%% Validate .json file

from data_management.databases import sanity_check_json_db

options = sanity_check_json_db.SanityCheckOptions()
options.baseDir = remote_image_base_dir
options.bCheckImageSizes = False
options.bCheckImageExistence = False
options.bFindUnusedImages = False

_, _, _ = sanity_check_json_db.sanity_check_json_db(output_json, options)


#%% Preview labels

from md_visualization import visualize_db

viz_options = visualize_db.DbVizOptions()
viz_options.num_to_visualize = 100
viz_options.trim_to_images_with_bboxes = False
viz_options.add_search_links = False
viz_options.sort_by_filename = False
viz_options.parallelize_rendering = True
viz_options.include_filename_links = True

# viz_options.classes_to_exclude = ['test']
html_output_file, _ = visualize_db.process_images(db_path=output_json,
                                                         output_dir=os.path.join(
                                                         base_folder,'preview'),
                                                         image_base_dir=remote_image_base_dir,
                                                         options=viz_options)
os.startfile(html_output_file)



########
#
# subset_json_detector_output.py
#
# Creates one or more subsets of a detector API output file (.json), doing either
# or both of the following (if both are requested, they happen in this order):
#
# 1) Retrieve all elements where filenames contain a specified query string, 
#    optionally replacing that query with a replacement token. If the query is blank, 
#    can also be used to prepend content to all filenames.
#
#    Does not support regex's, but supports a special case of ^string to indicate "must start with
#    to match".
#
# 2) Create separate .jsons for each unique path, optionally making the filenames 
#    in those .json's relative paths.  In this case, you specify an output directory, 
#    rather than an output path.  All images in the folder blah\foo\bar will end up 
#    in a .json file called blah_foo_bar.json.
#
###
#
# Sample invocations (splitting into multiple json's):
#
# Read from "1800_idfg_statewide_wolf_detections_w_classifications.json", split up into 
# individual .jsons in 'd:\temp\idfg\output', making filenames relative to their individual
# folders:
#
# python subset_json_detector_output.py "d:\temp\idfg\1800_idfg_statewide_wolf_detections_w_classifications.json" "d:\temp\idfg\output" --split_folders --make_folder_relative
#
# Now do the same thing, but instead of writing .json's to d:\temp\idfg\output, write them to *subfolders*
# corresponding to the subfolders for each .json file.
#
# python subset_json_detector_output.py "d:\temp\idfg\1800_detections_S2.json" "d:\temp\idfg\output_to_folders" --split_folders --make_folder_relative --copy_jsons_to_folders
#
###
#
# Sample invocations (creating a single subset matching a query):
#
# Read from "1800_detections.json", write to "1800_detections_2017.json"
#
# Include only images matching "2017", and change "2017" to "blah"
#
# python subset_json_detector_output.py "d:\temp\1800_detections.json" "d:\temp\1800_detections_2017_blah.json" --query 2017 --replacement blah
#
# Include all images, prepend with "prefix/"
#
# python subset_json_detector_output.py "d:\temp\1800_detections.json" "d:\temp\1800_detections_prefix.json" --replacement "prefix/"
#
###
#
# To subset a COCO Camera Traps .json database, see subset_json_db.py
#
########

#%% Constants and imports

import argparse
import sys
import copy
import json
import os
import re

from tqdm import tqdm

from md_utils.ct_utils import args_to_object
from md_utils.ct_utils import get_max_conf


#%% Helper classes

class SubsetJsonDetectorOutputOptions:
    
    # Only process files containing the token 'query'
    query = None
    
    # Replace 'query' with 'replacement' if 'replacement' is not None.  If 'query' is None,
    # prepend 'replacement'
    replacement = None
    
    # Should we split output into individual .json files for each folder?
    split_folders = False
    
    # Folder level to use for splitting ['bottom','top','n_from_bottom','n_from_top','dict']
    #
    # 'dict' requires 'split_folder_param' to be a dictionary mapping each filename
    # to a token.
    split_folder_mode = 'bottom'  # 'top'
    
    # When using the 'n_from_bottom' parameter to define folder splitting, this
    # defines the number of directories from the bottom.  'n_from_bottom' with
    # a parameter of zero is the same as 'bottom'.
    #
    # Same story with 'n_from_top'.
    #
    # When 'split_folder_mode' is 'dict', this should be a dictionary mapping each filename
    # to a token.
    split_folder_param = 0
    
    # Only meaningful if split_folders is True: should we convert pathnames to be relative
    # the folder for each .json file?
    make_folder_relative = False
    
    # Only meaningful if split_folders and make_folder_relative are True: if not None, 
    # will copy .json files to their corresponding output directories, relative to 
    # output_filename
    copy_jsons_to_folders = False
    
    # Should we over-write .json files?
    overwrite_json_files = False
    
    # If copy_jsons_to_folders is true, do we require that directories already exist?
    copy_jsons_to_folders_directories_must_exist = True
    
    # Threshold on confidence
    confidence_threshold = None
    
    # Should we remove failed images?
    remove_failed_images = False
    
    debug_max_images = -1
    
    
#%% Main function

def write_detection_results(data, output_filename, options):
    """
    Write the detector-output-formatted dict *data* to *output_filename*.
    """
    
    if (not options.overwrite_json_files) and os.path.isfile(output_filename):
        raise ValueError('File {} exists'.format(output_filename))
    
    basedir = os.path.dirname(output_filename)
    
    if options.copy_jsons_to_folders and options.copy_jsons_to_folders_directories_must_exist:
        if not os.path.isdir(basedir):
            raise ValueError('Directory {} does not exist'.format(basedir))
    else:
        os.makedirs(basedir, exist_ok=True)
    
    print('Writing detection output to {}'.format(output_filename))
    with open(output_filename, 'w') as f:
        json.dump(data,f,indent=1)

# ...write_detection_results()


def subset_json_detector_output_by_confidence(data, options):
    """
    Remove all detections below options.confidence_threshold, update max confidences accordingly.
    """
    
    if not options.confidence_threshold:
        return data
    
    images_in = data['images']
    images_out = []    
    
    print('Subsetting by confidence >= {}'.format(options.confidence_threshold))
    
    n_max_changes = 0
    
    # iImage = 0; im = images_in[0]
    for iImage, im in tqdm(enumerate(images_in), total=len(images_in)):
        
        if ('detections' not in im) or (im['detections'] is None):
            continue
        
        p_orig = get_max_conf(im)

        # Find all detections above threshold for this image
        detections = [d for d in im['detections'] if d['conf'] >= options.confidence_threshold]

        # If there are no detections above threshold, set the max probability
        # to -1, unless it already had a negative probability.
        if len(detections) == 0:
            if p_orig <= 0:                
                p = p_orig
            else:
                p = -1

        # Otherwise find the max confidence
        else:
            p = max(d['conf'] for d in detections)
        
        im['detections'] = detections

        # Did this thresholding result in a max-confidence change?
        if abs(p_orig - p) > 0.00001:

            # We should only be *lowering* max confidence values (i.e., making them negative)
            assert (p_orig <= 0) or (p < p_orig), 'Confidence changed from {} to {}'.format(p_orig, p)
            n_max_changes += 1
        
        if 'max_detection_conf' in im:
            im['max_detection_conf'] = p
        images_out.append(im)
        
    # ...for each image        
    
    data['images'] = images_out    
    print('done, found {} matches (of {}), {} max conf changes'.format(
            len(data['images']),len(images_in),n_max_changes))
    
    return data

# ...subset_json_detector_output_by_confidence()


def remove_failed_images(data,options):
    """
    Removed failed images from [data]
    """
    
    images_in = data['images']
    images_out = []    
    
    if not options.remove_failed_images:
        return data
        
    print('Removing failed images...', end='')
    
    # i_image = 0; im = images_in[0]
    for i_image, im in tqdm(enumerate(images_in), total=len(images_in)):
        
        if 'failure' in im and isinstance(im['failure'],str):
            continue
        else:
            images_out.append(im)
        
    # ...for each image        
    
    data['images'] = images_out    
    n_removed = len(images_in) - len(data['images'])
    print('Done, removed {} of {}'.format(n_removed, len(images_in)))
    
    return data

# ...remove_failed_images()


def subset_json_detector_output_by_query(data, options):
    """
    Subset to images whose filename matches options.query; replace all instances of 
    options.query with options.replacement.
    """
    
    images_in = data['images']
    images_out = []    
    
    print('Subsetting by query {}, replacement {}...'.format(options.query, options.replacement), end='')
    
    query_string = options.query
    query_starts_with = False
    
    # Support a special case regex-like notation for "starts with"
    if query_string is not None and query_string.startswith('^'):
        query_string = query_string[1:]
        query_starts_with = True
        
    # i_image = 0; im = images_in[0]
    for i_image, im in tqdm(enumerate(images_in), total=len(images_in)):
        
        fn = im['file']
        
        # Only take images that match the query
        if query_string is not None:
            if query_starts_with:
                if (not fn.startswith(query_string)):
                    continue
            else:
                if query_string not in fn:
                    continue
        
        if options.replacement is not None:
            if query_string is not None:
                fn = fn.replace(query_string, options.replacement)
            else:
                fn = options.replacement + fn
            
        im['file'] = fn
        
        images_out.append(im)
        
    # ...for each image        
    
    data['images'] = images_out    
    print('done, found {} matches (of {})'.format(len(data['images']), len(images_in)))
    
    return data

# ...subset_json_detector_output_by_query()


def split_path(path, maxdepth=100):
    """
    Splits [path] into all its constituent tokens, e.g.:
    
    c:\blah\boo\goo.txt
    
    ...becomes:
        
    ['c:\\', 'blah', 'boo', 'goo.txt']
    
    http://nicks-liquid-soapbox.blogspot.com/2011/03/splitting-path-to-list-in-python.html
    """
    
    (head, tail) = os.path.split(path)
    return split_path(head, maxdepth - 1) + [tail] \
        if maxdepth and head and head != path \
        else [head or tail]

# ...split_path()

    
def top_level_folder(p):
    """
    Gets the top-level folder from the path *p*; on Windows, will use the top-level folder
    that isn't the drive.  E.g., top_level_folder(r"c:\blah\foo") returns "c:\blah".  Does not
    include the leaf node, i.e. top_level_folder('/blah/foo') returns '/blah'.
    """
    
    if p == '':
        return ''
    
    # Path('/blah').parts is ('/','blah')
    parts = split_path(p)
    
    if len(parts) == 1:
        return parts[0]

    # Handle paths like:
    #
    # /, \, /stuff, c:, c:\stuff
    drive = os.path.splitdrive(p)[0]
    if parts[0] == drive or parts[0] == drive + '/' or parts[0] == drive + '\\' or parts[0] in ['\\', '/']:
        return os.path.join(parts[0], parts[1])
    else:
        return parts[0]

# ...top_level_folder()

    
if False:  
      
    p = 'blah/foo/bar'; s = top_level_folder(p); print(s); assert s == 'blah'
    p = '/blah/foo/bar'; s = top_level_folder(p); print(s); assert s == '/blah'
    p = 'bar'; s = top_level_folder(p); print(s); assert s == 'bar'
    p = ''; s = top_level_folder(p); print(s); assert s == ''
    p = 'c:\\'; s = top_level_folder(p); print(s); assert s == 'c:\\'
    p = r'c:\blah'; s = top_level_folder(p); print(s); assert s == 'c:\\blah'
    p = r'c:\foo'; s = top_level_folder(p); print(s); assert s == 'c:\\foo'
    p = r'c:/foo'; s = top_level_folder(p); print(s); assert s == 'c:/foo'
    p = r'c:\foo/bar'; s = top_level_folder(p); print(s); assert s == 'c:\\foo'
    
    
def subset_json_detector_output(input_filename, output_filename, options, data=None):
    """
    Main internal entry point
        
    Makes a copy of [data] before modifying if a data dictionary is supplied.
    """
    
    if options is None:    
        options = SubsetJsonDetectorOutputOptions()
            
    # Input validation        
    if options.copy_jsons_to_folders:
        assert options.split_folders and options.make_folder_relative, \
            'copy_jsons_to_folders set without make_folder_relative and split_folders'
                
    if options.split_folders:
        if os.path.isfile(output_filename):
            raise ValueError('When splitting by folders, output must be a valid directory name, you specified an existing file')
            
    if data is None:
        print('Reading json...', end='')
        with open(input_filename) as f:
            data = json.load(f)
        print(' ...done, read {} images'.format(len(data['images'])))
        if options.debug_max_images > 0:
            print('Trimming to {} images'.format(options.debug_max_images))
            data['images'] = data['images'][:options.debug_max_images]
    else:
        print('Copying data')
        data = copy.deepcopy(data)
        print('...done')
        
    if options.query is not None:
        
        data = subset_json_detector_output_by_query(data, options)
    
    if options.remove_failed_images:
        
        data = remove_failed_images(data, options)
        
    if options.confidence_threshold is not None:
        
        data = subset_json_detector_output_by_confidence(data, options)
        
    if not options.split_folders:
        
        write_detection_results(data, output_filename, options)
        return data
    
    else:
        
        # Map images to unique folders
        print('Finding unique folders')    
        
        folders_to_images = {}
        
        # im = data['images'][0]
        for im in tqdm(data['images']):
            
            fn = im['file']
            
            if options.split_folder_mode == 'bottom':
                                
                dirname = os.path.dirname(fn)
                
            elif options.split_folder_mode == 'n_from_bottom':
                
                dirname = os.path.dirname(fn)
                for n in range(0, options.split_folder_param):
                    dirname = os.path.dirname(dirname)
                    
            elif options.split_folder_mode == 'n_from_top':
                
                # Split string into folders, keeping delimiters
                
                # Don't use this, it removes delimiters
                # tokens = split_path(fn)
                tokens = re.split(r'([\\/])',fn)
                
                n_tokens_to_keep = ((options.split_folder_param + 1) * 2) - 1;
                
                if n_tokens_to_keep > len(tokens):
                    raise ValueError('Cannot walk {} folders from the top in path {}'.format(
                                options.split_folder_param, fn))
                dirname = ''.join(tokens[0:n_tokens_to_keep])
                
            elif options.split_folder_mode == 'top':
                
                dirname = top_level_folder(fn)                
                
            elif options.split_folder_mode == 'dict':
                
                assert isinstance(options.split_folder_param, dict)
                dirname = options.split_folder_param[fn]
                
            else:
                
                raise ValueError('Unrecognized folder split mode {}'.format(options.split_folder_mode))
                
            folders_to_images.setdefault(dirname, []).append(im)
        
        # ...for each image
                
        print('Found {} unique folders'.format(len(folders_to_images)))
        
        # Optionally make paths relative
        # dirname = list(folders_to_images.keys())[0]
        if options.make_folder_relative:
            
            print('Converting database-relative paths to individual-json-relative paths...')
        
            for dirname in tqdm(folders_to_images):
                # im = folders_to_images[dirname][0]
                for im in folders_to_images[dirname]:
                    fn = im['file']
                    relfn = os.path.relpath(fn, dirname).replace('\\', '/')
                    im['file'] = relfn
        
        # ...if we need to convert paths to be folder-relative
        
        print('Finished converting to json-relative paths, writing output')
                       
        os.makedirs(output_filename, exist_ok=True)
        all_images = data['images']
        
        # dirname = list(folders_to_images.keys())[0]
        for dirname in tqdm(folders_to_images):
                        
            json_fn = dirname.replace('/', '_').replace('\\', '_') + '.json'
            
            if options.copy_jsons_to_folders:
                json_fn = os.path.join(output_filename, dirname, json_fn)
            else:
                json_fn = os.path.join(output_filename, json_fn)
            
            # Recycle the 'data' struct, replacing 'images' every time... medium-hacky, but 
            # forward-compatible in that I don't take dependencies on the other fields
            dir_data = data
            dir_data['images'] = folders_to_images[dirname]
            write_detection_results(dir_data, json_fn, options)
            print('Wrote {} images to {}'.format(len(dir_data['images']), json_fn))
            
        # ...for each directory
        
        data['images'] = all_images
        
        return data
    
    # ...if we're splitting folders

# ...subset_json_detector_output()

    
#%% Interactive driver
                
if False:

    #%%
    
    #%% Subset a file without splitting
    
    input_filename = r"c:\temp\sample.json"
    output_filename = r"c:\temp\output.json"
     
    options = SubsetJsonDetectorOutputOptions()
    options.replacement = None
    options.query = 'S2'
        
    data = subset_json_detector_output(input_filename,output_filename,options,None)
    

    #%% Subset and split, but don't copy to individual folders

    input_filename = r"C:\temp\xxx-20201028_detections.filtered_rde_0.60_0.85_10_0.05_r2_export\xxx-20201028_detections.filtered_rde_0.60_0.85_10_0.05_r2_export.json"
    output_filename = r"c:\temp\out"
    
    options = SubsetJsonDetectorOutputOptions()
    options.split_folders = True    
    options.make_folder_relative = True
    options.split_folder_mode = 'n_from_top'
    options.split_folder_param = 1
    
    data = subset_json_detector_output(input_filename,output_filename,options,None)
    
    
    #%% Subset and split, copying to individual folders
    
    input_filename = r"c:\temp\sample.json"
    output_filename = r"c:\temp\out"
     
    options = SubsetJsonDetectorOutputOptions()
    options.split_folders = True    
    options.make_folder_relative = True
    options.copy_jsons_to_folders = True
    
    data = subset_json_detector_output(input_filename,output_filename,options,data)
    

#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument('input_file', type=str, help='Input .json filename')
    parser.add_argument('output_file', type=str, help='Output .json filename')
    parser.add_argument('--query', type=str, default=None, help='Query string to search for (omitting this matches all)')
    parser.add_argument('--replacement', type=str, default=None, help='Replace [query] with this')
    parser.add_argument('--confidence_threshold', type=float, default=None, help='Remove detections below this confidence level')
    parser.add_argument('--split_folders', action='store_true', help='Split .json files by leaf-node folder')
    parser.add_argument('--split_folder_param', type=int, help='Directory level count for n_from_bottom and n_from_top splitting')
    parser.add_argument('--split_folder_mode', type=str, help='Folder level to use for splitting ("top" or "bottom")')
    parser.add_argument('--make_folder_relative', action='store_true', help='Make image paths relative to their containing folder (only meaningful with split_folders)')
    parser.add_argument('--overwrite_json_files', action='store_true', help='Overwrite output files')
    parser.add_argument('--copy_jsons_to_folders', action='store_true', help='When using split_folders and make_folder_relative, copy jsons to their corresponding folders (relative to output_file)')
    parser.add_argument('--create_folders', action='store_true', help='When using copy_jsons_to_folders, create folders that don''t exist')    
    
    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()
        
    args = parser.parse_args()    
    
    # Convert to an options object
    options = SubsetJsonDetectorOutputOptions()
    if args.create_folders:
        options.copy_jsons_to_folders_directories_must_exist = False
        
    args_to_object(args, options)
    
    subset_json_detector_output(args.input_file, args.output_file, options)
    

if __name__ == '__main__':
    
    main()



########
#
# map_lila_taxonomy_to_wi_taxonomy.py
#
# Loads the LILA category mapping (in which taxonomy information comes from an 
# iNat taxonomy snapshot) and tries to map each class to the Wildlife Insights taxonomy.
#
########

#%% Constants and imports

import numpy as np
import json
import os

from tqdm import tqdm

from data_management.lila.lila_common import read_lila_taxonomy_mapping, \
    read_wildlife_insights_taxonomy_mapping

lila_local_base = os.path.expanduser('~/lila')

metadata_dir = os.path.join(lila_local_base, 'metadata')
os.makedirs(metadata_dir, exist_ok=True)

# Created by get_lila_category_list.py... contains counts for each category
category_list_dir = os.path.join(lila_local_base, 'lila_categories_list')
lila_dataset_to_categories_file = os.path.join(
    category_list_dir, 'lila_dataset_to_categories.json')

# This is a manually-curated file used to store mappings that had to be made manually
lila_to_wi_supplementary_mapping_file = os.path.expanduser(
    '~/git/MegaDetector/taxonomy_mapping/lila_to_wi_supplementary_mapping_file.csv')

assert os.path.isfile(lila_dataset_to_categories_file)

# This is the main output file from this whole process
wi_mapping_table_file = os.path.join(lila_local_base,'lila_wi_mapping_table.csv')

id_column = 'uniqueIdentifier' # 'id'


#%% Load category and taxonomy files

with open(lila_dataset_to_categories_file, 'r') as f:
    lila_dataset_to_categories = json.load(f)

lila_taxonomy_df = read_lila_taxonomy_mapping(metadata_dir)

wi_taxonomy_df = read_wildlife_insights_taxonomy_mapping(metadata_dir)


#%% Pull everything out of pandas

lila_taxonomy = lila_taxonomy_df.to_dict('records')
wi_taxonomy = wi_taxonomy_df.to_dict('records')


#%% Cache WI taxonomy lookups

def is_empty_wi_item(v):
    if isinstance(v, str):
        return len(v) == 0        
    elif v is None:
        return True
    else:
        assert isinstance(v, float) and np.isnan(v), 'Invalid item: {}'.format(str(v))
        return True


def taxonomy_items_equal(a, b):
    if isinstance(a, str) and (not isinstance(b, str)):
        return False
    if isinstance(b, str) and (not isinstance(a, str)):
        return False
    if (not isinstance(a, str)) or (not isinstance(b, str)):
        assert (a is None and b is None) or (isinstance(a, float) and isinstance(b, float))
        return True
    return a == b


for taxon in wi_taxonomy:
    taxon['taxon_name'] = None

from collections import defaultdict
wi_taxon_name_to_taxa = defaultdict(list)

# This is just a handy lookup table that we'll use to debug mismatches
wi_common_name_to_taxon = {}

blank_taxon_name = 'blank'
blank_taxon = None

animal_taxon_name = 'animal'
animal_taxon = None

unknown_taxon_name = 'unknown'
unknown_taxon = None

ignore_taxa = set(['No CV Result', 'CV Needed', 'CV Failed'])

known_problematic_taxon_ids = ['f94e6d97-59cf-4d38-a05a-a75efdd2863b']

human_taxa = []

# taxon = wi_taxonomy[21653]; print(taxon)
for taxon in tqdm(wi_taxonomy):

    taxon_name = None

    assert taxon['taxonomyType'] == 'object' or taxon['taxonomyType'] == 'biological'

    for k in taxon.keys():
        v = taxon[k]
        if isinstance(v,str):
            taxon[k] = v.strip()
            
    if taxon['commonNameEnglish'] in ignore_taxa:
        continue

    if isinstance(taxon['commonNameEnglish'], str):

        wi_common_name_to_taxon[taxon['commonNameEnglish'].strip(
        ).lower()] = taxon

        special_taxon = False

        # Look for keywords that don't refer to specific taxa: blank/animal/unknown
        if taxon['commonNameEnglish'].strip().lower() == blank_taxon_name:
            blank_taxon = taxon
            special_taxon = True

        elif taxon['commonNameEnglish'].strip().lower() == animal_taxon_name:
            animal_taxon = taxon
            special_taxon = True

        elif taxon['commonNameEnglish'].strip().lower() == unknown_taxon_name:
            unknown_taxon = taxon
            special_taxon = True

        if special_taxon:
            taxon_name = taxon['commonNameEnglish'].strip().lower()
            taxon['taxon_name'] = taxon_name
            wi_taxon_name_to_taxa[taxon_name].append(taxon)
            continue

    # Do we have a species name?
    if not is_empty_wi_item(taxon['species']):

        # If 'species' is populated, 'genus' should always be populated; one item currently breaks
        # this rule.
        assert not is_empty_wi_item(taxon['genus'])
        
        taxon_name = (taxon['genus'].strip() + ' ' +
                      taxon['species'].strip()).strip().lower()
        assert not is_empty_wi_item(taxon['class']) and \
            not is_empty_wi_item(taxon['order']) and \
            not is_empty_wi_item(taxon['family'])

    elif not is_empty_wi_item(taxon['genus']):

        assert not is_empty_wi_item(taxon['class']) and \
            not is_empty_wi_item(taxon['order']) and \
            not is_empty_wi_item(taxon['family'])
        taxon_name = taxon['genus'].strip().lower()

    elif not is_empty_wi_item(taxon['family']):

        assert not is_empty_wi_item(taxon['class']) and \
            not is_empty_wi_item(taxon['order'])
        taxon_name = taxon['family'].strip().lower()

    elif not is_empty_wi_item(taxon['order']):

        assert not is_empty_wi_item(taxon['class'])
        taxon_name = taxon['order'].strip().lower()

    elif not is_empty_wi_item(taxon['class']):

        taxon_name = taxon['class'].strip().lower()

    if taxon_name is not None:
        assert taxon['taxonomyType'] == 'biological'
    else:
        assert taxon['taxonomyType'] == 'object'
        taxon_name = taxon['commonNameEnglish'].strip().lower()

    if taxon_name in wi_taxon_name_to_taxa:
        if taxon[id_column] in known_problematic_taxon_ids:
            print('Skipping problematic taxon ID {}'.format(taxon[id_column]))
        else:
            previous_taxa = wi_taxon_name_to_taxa[taxon_name]
            for previous_taxon in previous_taxa:
                for level in ['class', 'order', 'family', 'genus', 'species']:
                    error_string = 'Error: taxon {} appeared previously in {} {} (as {}), now in {} {}'.format(
                        taxon_name,
                        level,previous_taxon[level],
                        previous_taxon['taxon_name'],
                        level,taxon[level])
                    assert taxonomy_items_equal(previous_taxon[level], taxon[level]), error_string
                
    taxon['taxon_name'] = taxon_name
    if taxon_name == 'homo sapiens':
        human_taxa.append(taxon)
    wi_taxon_name_to_taxa[taxon_name].append(taxon)

# ...for each taxon

assert unknown_taxon is not None
assert animal_taxon is not None
assert blank_taxon is not None


#%% Find redundant taxa

taxon_names_with_multiple_entries = []
for wi_taxon_name in wi_taxon_name_to_taxa:
    if len(wi_taxon_name_to_taxa[wi_taxon_name]) > 1:
        taxon_names_with_multiple_entries.append(wi_taxon_name)

print('{} names have multiple entries\n:'.format(len(taxon_names_with_multiple_entries)))

for s in taxon_names_with_multiple_entries:
    print(s)

if False:
    pass

    #%% Manual review of redundant taxa
    
    s = taxon_names_with_multiple_entries[15]
    taxa = wi_taxon_name_to_taxa[s]
    for t in taxa:
        for k in t.keys():
            print('{}: {}'.format(k,t[k]))
        print()
        # print(t,end='\n\n')


#%% Clean up redundant taxa

taxon_name_to_preferred_taxon_id = {}

# "helmeted guineafowl" vs "domestic guineafowl"
taxon_name_to_preferred_taxon_id['numida meleagris'] = '83133617-8358-4910-82ee-4c23e40ba3dc' # 2005826 

# "domestic turkey" vs. "wild turkey"
taxon_name_to_preferred_taxon_id['meleagris gallopavo'] = 'c10547c3-1748-48bf-a451-8066c820f22f' # 2021598 

# multiple sensible human entries
taxon_name_to_preferred_taxon_id['homo sapiens'] = '990ae9dd-7a59-4344-afcb-1b7b21368000' # 2002045 

# "domestic dog" and "dog-on-leash"
taxon_name_to_preferred_taxon_id['canis familiaris'] = '3d80f1d6-b1df-4966-9ff4-94053c7a902a' # 2021548 

# "small mammal" vs. "mammal"
taxon_name_to_preferred_taxon_id['mammalia'] = 'f2d233e3-80e3-433d-9687-e29ecc7a467a' # 2021108 

# "Hispaniolan Mango" vs. NaN
taxon_name_to_preferred_taxon_id['anthracothorax dominicus'] = 'f94e6d97-59cf-4d38-a05a-a75efdd2863b'

# "millipedes" vs. "Millipede"
taxon_name_to_preferred_taxon_id['diplopoda'] =  '065884eb-4e64-4233-84dc-de25bd06ffd2' # 2021760

# Different suborders: Squamata vs. Lacertilia
taxon_name_to_preferred_taxon_id['squamata'] = '710c4066-bd5d-4313-bcf4-0217c4c84da7' # 2021703

# Redundancy (both "beautiful firetail")
taxon_name_to_preferred_taxon_id['stagonopleura bella'] = '7fec8e7e-fd3b-4d7f-99fd-3ade6f3bbaa5' # 2021939

# "yellow wagtail" vs. "yellow crowned-wagtail"
taxon_name_to_preferred_taxon_id['motacilla flava'] = 'ac6669bc-9f9e-4473-b609-b9082f9bf50c' # 2016194 

# "dremomys species" vs. "dremomys genus"
taxon_name_to_preferred_taxon_id['dremomys'] = '1507d153-af11-46f1-bfb8-77918d035ab3' # 2019370

# "elk" vs. "domestic elk"
taxon_name_to_preferred_taxon_id['cervus canadensis'] = 'c5ce946f-8f0d-4379-992b-cc0982381f5e' 

# "American bison" vs. "domestic bison"
taxon_name_to_preferred_taxon_id['bison bison'] = '539ebd55-081b-429a-9ae6-5a6a0f6999d4' # 2021593 

# "woodrat or rat or mouse species" vs. "mouse species"
taxon_name_to_preferred_taxon_id['muridae'] = 'e7503287-468c-45af-a1bd-a17821bb62f2' # 2021642 

# both "southern sand frog"
taxon_name_to_preferred_taxon_id['tomopterna adiastola'] = 'a5dc63cb-41be-4090-84a7-b944b16dcee4' # 2021834

# sericornis species vs. scrubwren species
taxon_name_to_preferred_taxon_id['sericornis'] = 'ad82c0ac-df48-4028-bf71-d2b2f4bc4129' # 2021776

    
# taxon_name = list(taxon_name_to_preferred_taxon_id.keys())[0]
for taxon_name in taxon_name_to_preferred_taxon_id.keys():
    
    candidate_taxa = wi_taxon_name_to_taxa[taxon_name]
    
    # If we've gotten this far, we should be choosing from multiple taxa.
    #
    # This will become untrue if any of these are resolved later, at which point we should
    # remove them from taxon_name_to_preferred_id
    assert len(candidate_taxa) > 1, 'Only one taxon available for {}'.format(taxon_name)
    
    # Choose the preferred taxa
    selected_taxa = [t for t in candidate_taxa if t[id_column] == \
                     taxon_name_to_preferred_taxon_id[taxon_name]]
    assert len(selected_taxa) == 1
    wi_taxon_name_to_taxa[taxon_name] = selected_taxa

wi_taxon_name_to_taxon = {}

for taxon_name in wi_taxon_name_to_taxa.keys():
    taxa = wi_taxon_name_to_taxa[taxon_name]
    assert len(taxa) == 1
    wi_taxon_name_to_taxon[taxon_name] = taxa[0]


#%% Read supplementary mappings

with open(lila_to_wi_supplementary_mapping_file, 'r') as f:
    lines = f.readlines()

supplementary_lila_query_to_wi_query = {}

for line in lines:
    # Each line is [lila query],[WI taxon name],[notes]
    tokens = line.strip().split(',')
    assert len(tokens) == 3
    lila_query = tokens[0].strip().lower()
    wi_taxon_name = tokens[1].strip().lower()
    assert wi_taxon_name in wi_taxon_name_to_taxa
    supplementary_lila_query_to_wi_query[lila_query] = wi_taxon_name


#%% Map LILA categories to WI categories

mismatches = set()
mismatches_with_common_mappings = set()
supplementary_mappings = set()

all_searches = set()

# Must be ordered from kingdom --> species
lila_taxonomy_levels = ['kingdom', 'phylum', 'subphylum', 'superclass', 'class', 'subclass',
                        'infraclass', 'superorder', 'order', 'suborder', 'infraorder',
                        'superfamily', 'family', 'subfamily', 'tribe', 'genus', 'species']

unknown_queries = set(
    ['unidentifiable', 'other', 'unidentified', 'unknown', 'unclassifiable'])
blank_queries = set(['empty'])
animal_queries = set(['animalia'])

lila_dataset_category_to_wi_taxon = {}

# i_taxon = 0; taxon = lila_taxonomy[i_taxon]; print(taxon)
for i_taxon, lila_taxon in enumerate(lila_taxonomy):

    query = None

    lila_dataset_category = lila_taxon['dataset_name'] + ':' + lila_taxon['query']
    
    # Go from kingdom --> species, choosing the lowest-level description as the query
    for level in lila_taxonomy_levels:
        if isinstance(lila_taxon[level], str):
            query = lila_taxon[level]
            all_searches.add(query)

    if query is None:
        # E.g., 'car'
        query = lila_taxon['query']

    wi_taxon = None

    if query in unknown_queries:

        wi_taxon = unknown_taxon

    elif query in blank_queries:

        wi_taxon = blank_taxon

    elif query in animal_queries:

        wi_taxon = animal_taxon

    elif query in wi_taxon_name_to_taxon:

        wi_taxon = wi_taxon_name_to_taxon[query]

    elif query in supplementary_lila_query_to_wi_query:

        wi_taxon = wi_taxon_name_to_taxon[supplementary_lila_query_to_wi_query[query]]
        supplementary_mappings.add(query)
        # print('Made a supplementary mapping from {} to {}'.format(query,wi_taxon['taxon_name']))

    else:

        # print('No match for {}'.format(query))
        lila_common_name = lila_taxon['common_name']

        if lila_common_name in wi_common_name_to_taxon:
            wi_taxon = wi_common_name_to_taxon[lila_common_name]
            wi_common_name = wi_taxon['commonNameEnglish']
            wi_taxon_name = wi_taxon['taxon_name']
            if False:
                print('LILA common name {} maps to WI taxon {} ({})'.format(lila_common_name,
                                                                            wi_taxon_name,
                                                                            wi_common_name))
            mismatches_with_common_mappings.add(query)

        else:

            mismatches.add(query)

    lila_dataset_category_to_wi_taxon[lila_dataset_category] = wi_taxon

# ...for each LILA taxon

print('Of {} entities, there are {} mismatches ({} mapped by common name) ({} mapped by supplementary mapping file)'.format(
    len(all_searches), len(mismatches), len(mismatches_with_common_mappings), len(supplementary_mappings)))

assert len(mismatches) == 0


#%% Manual mapping

if not os.path.isfile(lila_to_wi_supplementary_mapping_file):
    print('Creating mapping file {}'.format(
        lila_to_wi_supplementary_mapping_file))
    with open(lila_to_wi_supplementary_mapping_file, 'w') as f:
        for query in mismatches:
            f.write(query + ',' + '\n')
else:
    print('{} exists, not re-writing'.format(lila_to_wi_supplementary_mapping_file))


#%% Build a dictionary from LILA dataset names and categories to LILA taxa

lila_dataset_category_to_lila_taxon = {}

# i_d = 0; d = lila_taxonomy[i_d]
for i_d,d in enumerate(lila_taxonomy):
    lila_dataset_category = d['dataset_name'] + ':' + d['query']
    assert lila_dataset_category not in lila_dataset_category_to_lila_taxon
    lila_dataset_category_to_lila_taxon[lila_dataset_category] = d


#%% Map LILA datasets to WI taxa, and count the number of each taxon available in each dataset

with open(wi_mapping_table_file,'w') as f:
    
    f.write('lila_dataset_name,lila_category_name,wi_guid,wi_taxon_name,wi_common,count\n')
    
    # dataset_name = list(lila_dataset_to_categories.keys())[0]
    for dataset_name in lila_dataset_to_categories.keys():
        
        if '_bbox' in dataset_name:
            continue
        
        dataset_categories = lila_dataset_to_categories[dataset_name]
        
        # dataset_category = dataset_categories[0]
        for category in dataset_categories:
            
            lila_dataset_category = dataset_name + ':' + category['name'].strip().lower()
            if '#' in lila_dataset_category:
                continue
            assert lila_dataset_category in lila_dataset_category_to_lila_taxon
            assert lila_dataset_category in lila_dataset_category_to_wi_taxon
            assert 'count' in category
    
            wi_taxon = lila_dataset_category_to_wi_taxon[lila_dataset_category]
            
            # Write out the dataset name, category name, WI GUID, WI scientific name, WI common name, 
            # and count
            s = f"{dataset_name},{category['name']},{wi_taxon['uniqueIdentifier']},"+\
                f"{wi_taxon['taxon_name']},{wi_taxon['commonNameEnglish']},{category['count']}\n"
            f.write(s)
            
        # ...for each category in this dataset
            
    # ...for each dataset    

# ...with open()


########
#
# visualize_megadb.py
#
# Create visual previews of images/sequences in MegaDB.
#
########

#%% Imports

import argparse
import json
import os
import sys
from random import shuffle
from multiprocessing.pool import ThreadPool
from functools import partial
import io

from tqdm import tqdm

from data_management.megadb.megadb_utils import MegadbUtils
from md_utils.write_html_image_list import write_html_image_list
from md_visualization import visualization_utils as vis_utils


#%% Support functions

def render_image_info(rendering, args):
    
    storage_client = rendering['storage_client']
    image_obj = io.BytesIO()

    try:
        storage_client.download_blob(rendering['blob_path']).readinto(image_obj)
    except Exception as e:
        print(f'Image not found in blob storage: {rendering["blob_path"]}')
        print(e)
        return

    # resize is for displaying them more quickly
    image = vis_utils.resize_image(
        vis_utils.open_image(image_obj), args.output_image_width)
    vis_utils.render_megadb_bounding_boxes(rendering['bbox'], image)

    annotated_img_name = rendering['annotated_img_name']
    annotated_img_path = os.path.join(
        args.output_dir, 'rendered_images', annotated_img_name)
    image.save(annotated_img_path)


def visualize_sequences(datasets_table, sequences, args):
    
    num_images = 0

    images_html = []
    rendering_info = []

    for seq in sequences:
        if 'images' not in seq:
            continue

        # dataset and seq_id are required fields
        dataset_name = seq['dataset']
        seq_id = seq['seq_id']

        # sort the images in the sequence

        images_in_seq = sorted(seq['images'], key=lambda x: x['frame_num']) if len(seq['images']) > 1 else seq['images']

        for im in images_in_seq:
            if args.trim_to_images_bboxes_labeled and 'bbox' not in im:
                continue

            num_images += 1

            blob_path = MegadbUtils.get_full_path(
                datasets_table, dataset_name, im['file'])
            frame_num = im.get('frame_num', -1)

            # if no class label on the image, show class label on the sequence
            im_class = im.get('class', None)
            if im_class is None:
                im_class = seq.get('class', [])

            rendering = {}
            rendering['storage_client'] = MegadbUtils.get_storage_client(
                datasets_table, dataset_name)
            rendering['blob_path'] = blob_path
            rendering['bbox'] = im.get('bbox', [])

            annotated_img_name = 'anno_' + blob_path.replace('/', args.pathsep_replacement).replace('\\', args.pathsep_replacement)
            rendering['annotated_img_name'] = annotated_img_name

            rendering_info.append(rendering)

            images_html.append({
                'filename': 'rendered_images/{}'.format(annotated_img_name),
                'title': 'Seq ID: {}. Frame number: {}<br/> Image file: {}<br/> number of boxes: {}, image class labels: {}'.format(seq_id, frame_num, blob_path, len(rendering['bbox']), im_class),
                'textStyle': 'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5'
            })

        if num_images >= args.num_to_visualize:
            print('num_images visualized is {}'.format(num_images))
            break

    # pool = ThreadPool()
    render_image_info_partial = partial(render_image_info, args=args)
    # print('len of rendering_info', len(rendering_info))
    # tqdm(pool.imap_unordered(render_image_info_partial, rendering_info), total=len(rendering_info))

    for rendering in tqdm(rendering_info):
        render_image_info_partial(rendering)

    print('Making HTML...')

    html_path = os.path.join(args.output_dir, 'index.html')
    # options = write_html_image_list()
    # options['headerHtml']
    write_html_image_list(
        filename=html_path,
        images=images_html
    )


#%% Command-line driver

def main():
    
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'megadb_entries', type=str,
        help='Path to a json list of MegaDB entries')
    parser.add_argument(
        'output_dir', action='store', type=str,
        help='Output directory for html and rendered images')
    parser.add_argument(
        '--trim_to_images_bboxes_labeled', action='store_true',
        help='Only include images that have been sent for bbox labeling (but '
             'may be actually empty). Turn this on if QAing annotations.')
    parser.add_argument(
        '--num_to_visualize', action='store', type=int, default=200,
        help='Number of images to visualize (all comformant images in a '
             'sequence are shown, so may be a few more than specified). '
             'Sequences are shuffled. Default: 200. Use -1 to visualize all.')
    parser.add_argument(
        '--pathsep_replacement', action='store', type=str, default='~',
        help='Replace path separators in relative filenames with another '
             'character (default ~)')
    parser.add_argument(
        '-w', '--output_image_width', type=int, default=700,
        help='an integer indicating the desired width in pixels of the output '
             'annotated images. Use -1 to not resize.')

    if len(sys.argv[1:]) == 0:
        parser.print_help()
        parser.exit()

    args = parser.parse_args()

    assert 'COSMOS_ENDPOINT' in os.environ and 'COSMOS_KEY' in os.environ

    os.makedirs(args.output_dir, exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, 'rendered_images'))

    print('Connecting to MegaDB to get the datasets table...')
    megadb_utils = MegadbUtils()
    datasets_table = megadb_utils.get_datasets_table()

    print('Loading the MegaDB entries...')
    with open(args.megadb_entries) as f:
        sequences = json.load(f)
    print('Total number of sequences: {}'.format(len(sequences)))

    # print('Checking that the MegaDB entries conform to the schema...')
    # sequences_schema_check.sequences_schema_check(sequences)

    shuffle(sequences)
    visualize_sequences(datasets_table, sequences, args)


if __name__ == '__main__':
    main()



########
#
# coco_to_yolo.py
#
# Converts a COCO-formatted dataset to a YOLO-formatted dataset. 
#
# If the input and output folders are the same, writes .txt files to the input folder,
# and neither moves nor modifies images.
#
# Currently ignores segmentation masks, and errors if an annotation has a 
# segmentation polygon but no bbox
# 
# Has only been tested on a handful of COCO Camera Traps data sets; if you
# use it for more general COCO conversion, YMMV.
#
########

#%% Imports and constants

import json
import os
import shutil

from collections import defaultdict
from tqdm import tqdm


#%% Support functions

def coco_to_yolo(input_image_folder,output_folder,input_file,
                 source_format='coco',overwrite_images=False,
                 create_image_and_label_folders=False,
                 class_file_name='classes.txt',
                 allow_empty_annotations=False,
                 clip_boxes=False,
                 image_id_to_output_image_json_file=None):
    
    if output_folder is None:
        output_folder = input_image_folder
    
    # Validate input
    
    assert os.path.isdir(input_image_folder)
    assert os.path.isfile(input_file)
    os.makedirs(output_folder,exist_ok=True)
    
    
    # Read input data
    
    with open(input_file,'r') as f:
        data = json.load(f)
        
        
    # Parse annotations
  
    image_id_to_annotations = defaultdict(list)
    
    # i_ann = 0; ann = data['annotations'][0]
    for i_ann,ann in enumerate(data['annotations']):
        
        # Make sure no annotations have *only* segmentation data 
        if ( \
            ('segmentation' in ann.keys()) and \
            (ann['segmentation'] is not None) and \
            (len(ann['segmentation']) > 0) ) \
            and \
            (('bbox' not in ann.keys()) or (ann['bbox'] is None) or (len(ann['bbox'])==0)):
                raise ValueError('Oops: segmentation data present without bbox information, ' + \
                                 'this script isn\'t ready for this dataset')
        
        image_id_to_annotations[ann['image_id']].append(ann)
        
    print('Parsed annotations for {} images'.format(len(image_id_to_annotations)))
        
    # Re-map class IDs to make sure they run from 0...n-classes-1
    #
    # TODO: this allows unused categories in the output data set, which I *think* is OK,
    # but I'm only 81% sure.
    next_category_id = 0
    coco_id_to_yolo_id = {}
    yolo_id_to_name = {}
    for category in data['categories']:
        assert category['id'] not in coco_id_to_yolo_id
        coco_id_to_yolo_id[category['id']] = next_category_id
        yolo_id_to_name[next_category_id] = category['name']
        next_category_id += 1
        
    
    # Process images (everything but I/O)
    
    # List of dictionaries with keys 'source_image','dest_image','bboxes','dest_txt'
    images_to_copy = []
    
    missing_images = []
    
    image_names = set()
    
    typical_image_extensions = set(['.jpg','.jpeg','.png','.gif','.tif','.bmp'])
    
    printing_empty_annotation_warning = False
    
    image_id_to_output_image_name = {}
    
    # i_image = 0; im = data['images'][i_image]
    for i_image,im in tqdm(enumerate(data['images']),total=len(data['images'])):
                
        output_info = {}
        source_image = os.path.join(input_image_folder,im['file_name'])        
        output_info['source_image'] = source_image
        
        tokens = os.path.splitext(im['file_name'])
        if tokens[1].lower() not in typical_image_extensions:
            print('Warning: unusual image file name {}'.format(im['file_name']))
                  
        image_name = tokens[0].replace('\\','/').replace('/','_') + '_' + str(i_image).zfill(6)
        assert image_name not in image_names, 'Image name collision for {}'.format(image_name)
        image_names.add(image_name)
        
        assert im['id'] not in image_id_to_output_image_name
        image_id_to_output_image_name[im['id']] = image_name
        
        dest_image_relative = image_name + tokens[1]
        output_info['dest_image_relative'] = dest_image_relative
        dest_txt_relative = image_name + '.txt'
        output_info['dest_txt_relative'] = dest_txt_relative
        output_info['bboxes'] = []
        
        # assert os.path.isfile(source_image), 'Could not find image {}'.format(source_image)
        if not os.path.isfile(source_image):
            print('Warning: could not find image {}'.format(source_image))
            missing_images.append(im['file_name'])
            continue
        
        image_id = im['id']
        
        image_bboxes = []
                
        if image_id in image_id_to_annotations:
                        
            for ann in image_id_to_annotations[image_id]:
                
                # If this annotation has no bounding boxes...
                if 'bbox' not in ann or ann['bbox'] is None or len(ann['bbox']) == 0:
                    
                    if source_format == 'coco':
                    
                        if not allow_empty_annotations:
                            # This is not entirely clear from the COCO spec, but it seems to be consensus
                            # that if you want to specify an image with no objects, you don't include any
                            # annotations for that image.
                            raise ValueError('If an annotation exists, it should have content')
                        else:
                            continue
                    
                    else:
                        
                        # We allow empty bbox lists in COCO camera traps; this is typically a negative
                        # example in a dataset that has bounding boxes, and 0 is typically the empty 
                        # category.
                        if ann['category_id'] != 0:
                            if not printing_empty_annotation_warning:
                                printing_empty_annotation_warning = True
                                print('Warning: empty annotation found with category {}'.format(
                                    ann['category_id']))
                        continue
                    
                # ...if this is an empty annotation
                
                coco_bbox = ann['bbox']
                yolo_category_id = coco_id_to_yolo_id[ann['category_id']]
                
                # COCO: [x_min, y_min, width, height] in absolute coordinates
                # YOLO: [class, x_center, y_center, width, height] in normalized coordinates
                
                # Convert from COCO coordinates to YOLO coordinates
                img_w = im['width']
                img_h = im['height']
                                
                if source_format == 'coco':
                    
                    x_min_absolute = coco_bbox[0]
                    y_min_absolute = coco_bbox[1]
                    box_w_absolute = coco_bbox[2]
                    box_h_absolute = coco_bbox[3]
                    
                    x_center_absolute = (x_min_absolute + (x_min_absolute + box_w_absolute)) / 2
                    y_center_absolute = (y_min_absolute + (y_min_absolute + box_h_absolute)) / 2
                    
                    x_center_relative = x_center_absolute / img_w
                    y_center_relative = y_center_absolute / img_h
                    
                    box_w_relative = box_w_absolute / img_w
                    box_h_relative = box_h_absolute / img_h
                    
                else:
                    
                    raise ValueError('Unrecognized source format {}'.format(source_format))
                
                if clip_boxes:
                    
                    clipped_box = False
                    
                    box_right = x_center_relative + (box_w_relative / 2.0)                    
                    if box_right > 1.0:
                        clipped_box = True
                        overhang = box_right - 1.0
                        box_w_relative -= overhang
                        x_center_relative -= (overhang / 2.0)

                    box_bottom = y_center_relative + (box_h_relative / 2.0)                                        
                    if box_bottom > 1.0:
                        clipped_box = True
                        overhang = box_bottom - 1.0
                        box_h_relative -= overhang
                        y_center_relative -= (overhang / 2.0)
                    
                    box_left = x_center_relative - (box_w_relative / 2.0)
                    if box_left < 0.0:
                        clipped_box = True
                        overhang = abs(box_left)
                        box_w_relative -= overhang
                        x_center_relative += (overhang / 2.0)
                        
                    box_top = y_center_relative - (box_h_relative / 2.0)
                    if box_top < 0.0:
                        clipped_box = True
                        overhang = abs(box_top)
                        box_h_relative -= overhang
                        y_center_relative += (overhang / 2.0)
                        
                    if clipped_box:
                        print('Warning: clipped box for image {}'.format(image_id))
                
                yolo_box = [yolo_category_id,
                            x_center_relative, y_center_relative, 
                            box_w_relative, box_h_relative]
                
                image_bboxes.append(yolo_box)
                
            # ...for each annotation 
            
        # ...if this image has annotations
        
        output_info['bboxes'] = image_bboxes
        
        images_to_copy.append(output_info)        
    
    # ...for each image
        
    print('{} missing images (of {})'.format(len(missing_images),len(data['images'])))
    
    
    # Write output
    
    print('Generating class list')
    
    class_list_filename = os.path.join(output_folder,class_file_name)
    with open(class_list_filename, 'w') as f:
        print('Writing class list to {}'.format(class_list_filename))
        for i_class in range(0,len(yolo_id_to_name)):
            # Category IDs should range from 0..N-1
            assert i_class in yolo_id_to_name
            f.write(yolo_id_to_name[i_class] + '\n')
    
    if image_id_to_output_image_json_file is not None:
        print('Writing image ID mapping to {}'.format(image_id_to_output_image_json_file))
        with open(image_id_to_output_image_json_file,'w') as f:
            json.dump(image_id_to_output_image_name,f,indent=1)
            
    print('Copying images and creating annotation files')
    
    if create_image_and_label_folders:
        dest_image_folder = os.path.join(output_folder,'images')
        dest_txt_folder = os.path.join(output_folder,'labels')
    else:
        dest_image_folder = output_folder
        dest_txt_folder = output_folder
        
    # TODO: parallelize this loop
    #
    # output_info = images_to_copy[0]
    for output_info in tqdm(images_to_copy):

        source_image = output_info['source_image']
        dest_image_relative = output_info['dest_image_relative']
        dest_txt_relative = output_info['dest_txt_relative']
        
        dest_image = os.path.join(dest_image_folder,dest_image_relative)
        os.makedirs(os.path.dirname(dest_image),exist_ok=True)
        
        dest_txt = os.path.join(dest_txt_folder,dest_txt_relative)
        os.makedirs(os.path.dirname(dest_txt),exist_ok=True)
        
        if not create_image_and_label_folders:
            assert os.path.dirname(dest_image) == os.path.dirname(dest_txt)
        
        if (not os.path.isfile(dest_image)) or (overwrite_images):
            shutil.copyfile(source_image,dest_image)
        
        bboxes = output_info['bboxes']        
        
        # Only write an annotation file if there are bounding boxes.  Images with 
        # no .txt files are treated as hard negatives, at least by YOLOv5:
        #
        # https://github.com/ultralytics/yolov5/issues/3218
        #
        # I think this is also true for images with empty annotation files, but 
        # I'm using the convention suggested on that issue, i.e. hard negatives 
        # are expressed as images without .txt files.
        if len(bboxes) > 0:
            
            with open(dest_txt,'w') as f:
                
                # bbox = bboxes[0]
                for bbox in bboxes:
                    assert len(bbox) == 5
                    s = '{} {} {} {} {}'.format(bbox[0],bbox[1],bbox[2],bbox[3],bbox[4])
                    f.write(s + '\n')
            
    # ...for each image              

# ...def coco_to_yolo()


#%% Interactive driver

if False:
    
    pass

    #%% CCT data
    
    input_image_folder = os.path.expanduser('~/data/noaa-fish/JPEGImages')
    output_folder = os.path.expanduser('~/data/noaa-fish/AllImagesWithAnnotations')
    input_file = os.path.expanduser('~/data/noaa-fish/noaa_estuary_fish.json')

    # If preview_export is True, I'm exporting to preview these with BoundingBoxEditor:
    #
    # https://github.com/mfl28/BoundingBoxEditor
    #
    # This export will be compatible, other than the fact that you need to move
    # "object.data" into the "labels" folder.
    #
    # Otherwise I'm exporting for training, in the YOLOv5 flat format.
    preview_export = False
    
    if preview_export:
        
        coco_to_yolo(input_image_folder,output_folder,input_file,
                     source_format='coco',
                     overwrite_images=False,
                     create_image_and_label_folders=True,
                     class_file_name='object.data',
                     allow_empty_annotations=True,
                     clip_boxes=True)
        
    else:
        
        coco_to_yolo(input_image_folder,output_folder,input_file,
                     source_format='coco',
                     overwrite_images=False,
                     create_image_and_label_folders=False,
                     class_file_name='classes.txt',
                     allow_empty_annotations=True,
                     clip_boxes=True,
                     image_id_to_output_image_json_file=\
                         os.path.join(output_folder,'image_id_to_output_image_name.json'))
    
    

#%% Command-line driver

# TODO



########
#
# remove_exif.py
#
# Removes all EXIF/IPTC/XMP metadata from a folder of images, without making 
# backup copies, using pyexiv2.
#
########

#%% Imports and constants

import os
import glob

input_base = r'f:\images'
assert os.path.isdir(input_base)


#%% List files

all_files = [f for f in glob.glob(input_base + "*/**", recursive=True)]
image_files = [s for s in all_files if (s.lower().endswith('.jpg'))]
    

#%% Remove EXIF data (support)

import pyexiv2

# PYEXIV2 IS NOT THREAD SAFE; DO NOT CALL THIS IN PARALLEL FROM A SINGLE PROCESS
def remove_exif(fn):
    
    try:
        img = pyexiv2.Image(fn)
        # data = img.read_exif(); print(data)
        img.clear_exif()
        img.clear_iptc()
        img.clear_xmp()
        img.close()        
    except Exception as e:
        print('EXIF error on {}: {}'.format(fn,str(e)))
    

#%% Debug

if False:    
    #%%
    fn = image_files[-10001]
    os.startfile(fn)    
    #%%
    remove_exif(fn)
    os.startfile(fn)
    
    
#%% Remove EXIF data (execution)

from joblib import Parallel, delayed

n_exif_threads = 50
    
if n_exif_threads == 1:
    
    # fn = image_files[0]
    for fn in image_files:
        remove_exif(fn)
        
else:
    # joblib.Parallel defaults to a process-based backend, but let's be sure
    # results = Parallel(n_jobs=n_exif_threads,verbose=2,prefer='processes')(delayed(remove_exif)(fn) for fn in image_files[0:10])
    results = Parallel(n_jobs=n_exif_threads,verbose=2,prefer='processes')(delayed(remove_exif)(fn) for fn in image_files)




########
#
# lila_common.py
#
# Common constants and functions related to LILA data management/retrieval.
#
########

#%% Imports and constants

import os
import json
import zipfile
import pandas as pd
import numpy as np

from urllib.parse import urlparse

from md_utils.url_utils import download_url
from md_utils.path_utils import unzip_file

# LILA camera trap primary metadata file
lila_metadata_url = 'http://lila.science/wp-content/uploads/2023/06/lila_camera_trap_datasets.csv'
lila_taxonomy_mapping_url = 'https://lila.science/wp-content/uploads/2022/07/lila-taxonomy-mapping_release.csv'
lila_all_images_url = 'https://lila.science/public/lila_image_urls_and_labels.csv.zip'

wildlife_insights_page_size = 30000
wildlife_insights_taxonomy_url = 'https://api.wildlifeinsights.org/api/v1/taxonomy/taxonomies-all?fields=class,order,family,genus,species,authority,taxonomyType,uniqueIdentifier,commonNameEnglish&page[size]={}'.format(
    wildlife_insights_page_size)
wildlife_insights_taxonomy_local_json_filename = 'wi_taxonomy.json'
wildlife_insights_taxonomy_local_csv_filename = \
    wildlife_insights_taxonomy_local_json_filename.replace('.json','.csv')

lila_azure_storage_account = 'https://lilablobssc.blob.core.windows.net'
gcp_bucket_api_url = 'https://storage.googleapis.com/public-datasets-lila'
gcp_bucket_gs_url = 'gs://public-datasets-lila'


#%% Common functions

def read_wildlife_insights_taxonomy_mapping(metadata_dir):
    """
    Reads the WI taxonomy mapping file, downloading the .json data (and writing to .csv) if necessary.
    
    Returns a Pandas dataframe.
    """
    
    wi_taxonomy_csv_path = os.path.join(metadata_dir,wildlife_insights_taxonomy_local_csv_filename)
    
    if os.path.exists(wi_taxonomy_csv_path):
        df = pd.read_csv(wi_taxonomy_csv_path)
    else:
        wi_taxonomy_json_path = os.path.join(metadata_dir,wildlife_insights_taxonomy_local_json_filename)
        download_url(wildlife_insights_taxonomy_url, wi_taxonomy_json_path)
        with open(wi_taxonomy_json_path,'r') as f:
            d = json.load(f)
            
        # We haven't implemented paging, make sure that's not an issue
        assert d['meta']['totalItems'] < wildlife_insights_page_size
            
        # d['data'] is a list of items that look like:
        """
         {'id': 2000003,
         'class': 'Mammalia',
         'order': 'Rodentia',
         'family': 'Abrocomidae',
         'genus': 'Abrocoma',
         'species': 'bennettii',
         'authority': 'Waterhouse, 1837',
         'commonNameEnglish': "Bennett's Chinchilla Rat",
         'taxonomyType': 'biological',
         'uniqueIdentifier': '7a6c93a5-bdf7-4182-82f9-7a67d23f7fe1'}
        """
        df = pd.DataFrame(d['data'])
        df.to_csv(wi_taxonomy_csv_path,index=False)
        
    return df

    
def read_lila_taxonomy_mapping(metadata_dir):
    """
    Reads the LILA taxonomy mapping file, downloading the .csv file if necessary.
    
    Returns a Pandas dataframe, with one row per identification.
    """
    
    p = urlparse(lila_taxonomy_mapping_url)
    taxonomy_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_taxonomy_mapping_url, taxonomy_filename)
    
    df = pd.read_csv(lila_taxonomy_mapping_url)
    
    return df

   
def is_empty(v):
    if v is None:
        return True
    if isinstance(v,str) and v == '':
        return True
    if isinstance(v,float) and np.isnan(v):
        return True
    return False


def read_lila_metadata(metadata_dir):
    """
    Reads LILA metadata (URLs to each dataset), downloading the txt file if necessary.
    
    Returns a dict mapping dataset names (e.g. "Caltech Camera Traps") to dicts
    with keys corresponding to the headers in the .csv file, currently:
        
    name,image_base_url,metadata_url,bbox_url,continent,country,region
    """
    
    # Put the master metadata file in the same folder where we're putting images
    p = urlparse(lila_metadata_url)
    metadata_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_metadata_url, metadata_filename)
    
    df = pd.read_csv(metadata_filename)
    
    records = df.to_dict('records')
    
    # Parse into a table keyed by dataset name
    metadata_table = {}
    
    # r = records[0]
    for r in records:
        if is_empty(r['name']):
            continue
        
        # Convert NaN's to None
        for k in r.keys():
            if is_empty(r[k]):
                r[k] = None
                
        metadata_table[r['name']] = r
    
    return metadata_table    
    

def read_lila_all_images_file(metadata_dir):
    """
    Downloads if necessary - then unzips if necessary - the .csv file with label mappings for
    all LILA files, and opens the resulting .csv file as a Pandas DataFrame.
    """
        
    p = urlparse(lila_all_images_url)
    lila_all_images_zip_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(lila_all_images_url, lila_all_images_zip_filename)
    
    with zipfile.ZipFile(lila_all_images_zip_filename,'r') as z:
        files = z.namelist()
    assert len(files) == 1
    
    unzipped_csv_filename = os.path.join(metadata_dir,files[0])
    if not os.path.isfile(unzipped_csv_filename):
        unzip_file(lila_all_images_zip_filename,metadata_dir)
    else:
        print('{} already unzipped'.format(unzipped_csv_filename))    
    
    df = pd.read_csv(unzipped_csv_filename)
    
    return df


def read_metadata_file_for_dataset(ds_name,metadata_dir,metadata_table=None):
    """
    Downloads if necessary - then unzips if necessary - the .json file for a specific dataset.
    Returns the .json filename on the local disk.
    """
    
    if metadata_table is None:
        metadata_table = read_lila_metadata(metadata_dir)
        
    json_url = metadata_table[ds_name]['metadata_url']
    
    p = urlparse(json_url)
    json_filename = os.path.join(metadata_dir,os.path.basename(p.path))
    download_url(json_url, json_filename)
    
    # Unzip if necessary
    if json_filename.endswith('.zip'):
        
        with zipfile.ZipFile(json_filename,'r') as z:
            files = z.namelist()
        assert len(files) == 1
        unzipped_json_filename = os.path.join(metadata_dir,files[0])
        if not os.path.isfile(unzipped_json_filename):
            unzip_file(json_filename,metadata_dir)        
        else:
            print('{} already unzipped'.format(unzipped_json_filename))
        json_filename = unzipped_json_filename
    
    return json_filename


def azure_url_to_gcp_http_url(url):
    """
    Most URLs point to Azure by default, but most files are available on both Azure and GCP.
    This function converts an Azure URL to the corresponding GCP http:// url.
    """
    
    assert url.startswith(lila_azure_storage_account)
    gcp_url = url.replace(lila_azure_storage_account,gcp_bucket_api_url,1)
    return gcp_url


def azure_url_to_gcp_gs_url(url):
    """
    Most URLs point to Azure by default, but most files are available on both Azure and GCP.
    This function converts an Azure URL to the corresponding GCP gs:// url.
    """
    
    return azure_url_to_gcp_http_url(url).replace(gcp_bucket_api_url,
                                                  gcp_bucket_gs_url,1)


#%% Interactive test driver

if False:
    
    pass

    #%% Verify that all base URLs exist
    
    # LILA camera trap primary metadata file
    urls = (lila_metadata_url,lila_taxonomy_mapping_url,lila_all_images_url,wildlife_insights_taxonomy_url)
    
    from md_utils import url_utils
    
    status_codes = url_utils.test_urls(urls)
    
    
    #%% Verify that the metadata URLs exist for individual datasets
    
    metadata_dir = os.path.expanduser('~/lila/metadata')
    
    dataset_metadata = read_lila_metadata(metadata_dir)
    
    urls_to_test = []
    # ds_name = next(iter(dataset_metadata.keys()))
    for ds_name in dataset_metadata.keys():
        
        ds_info = dataset_metadata[ds_name]
        urls_to_test.append(ds_info['metadata_url'])
        if ds_info['bbox_url'] != None:
            urls_to_test.append(ds_info['bbox_url'])
            
    status_codes = url_utils.test_urls(urls_to_test)    
    
    
    #%% Verify that the GCP versions of all metadata files exist
    
    gcp_urls = []
    
    # url = urls_to_test[0]
    for url in urls_to_test:
        assert url.startswith(lila_azure_storage_account)
        gcp_url = url.replace(lila_azure_storage_account,gcp_bucket_api_url,1)
        gcp_urls.append(gcp_url)
        
    status_codes = url_utils.test_urls(gcp_urls)


########
#
# visualize_incoming_annotations.py
#
# Spot-check the annotations received from iMerit by visualizing annotated bounding 
# boxes on a sample of images and display them in HTML.
#
# Modified in 2021 March to use the new format (iMerit batch 12 onwards), which is a
# COCO formatted JSON with relative coordinates for the bbox.
#
########

import argparse
import io
import json
import os

from collections import defaultdict
from random import sample

from tqdm import tqdm

from md_utils.write_html_image_list import write_html_image_list
from data_management.megadb.megadb_utils import MegadbUtils
from data_management.annotations.add_bounding_boxes_to_megadb import file_name_to_parts
from data_management.cct_json_utils import IndexedJsonDb
from md_visualization import visualization_utils as vis_utils


def get_image_rel_path(dataset_seq_images, dataset_name, seq_id, frame_num):
    images = dataset_seq_images[dataset_name].get(seq_id, None)
    if images is None:
        return None
    for im in images:
        if im.get('frame_num', None) == frame_num:
            return im['file']
        
    # we used frame_num of 1 when sending out images to annotators when it is not explicitly stored (wcs esp.)
    if frame_num == 1 and len(images) == 1:
        return images[0]['file']

    return None


def visualize_incoming_annotations(args):
    print('Connecting to MegaDB to get the datasets table...')
    megadb_utils = MegadbUtils()
    datasets_table = megadb_utils.get_datasets_table()

    print('Loading the MegaDB entries...')
    with open(args.megadb_entries) as f:
        sequences = json.load(f)
    print(f'Total number of sequences: {len(sequences)}')
    dataset_seq_images = defaultdict(dict)
    for seq in sequences:
        dataset_seq_images[seq['dataset']][seq['seq_id']] = seq['images']

    print('Loading incoming annotation entries...')
    incoming = IndexedJsonDb(args.incoming_annotation)
    print(f'Number of images in this annotation file: {len(incoming.image_id_to_image)}')

    if args.num_to_visualize != -1 and args.num_to_visualize <= len(incoming.image_id_to_image):
        incoming_id_to_anno = sample(list(incoming.image_id_to_annotations.items()),
                                     args.num_to_visualize)
    else:
        incoming_id_to_anno = incoming.image_id_to_annotations.items()

    # The file_name field in the incoming json looks like
    # alka_squirrels.seq2020_05_07_25C.frame119221.jpg
    # we need to use the dataset, sequence and frame info to find the actual path in blob storage
    # using the sequences
    images_html = []
    for image_id, annotations in tqdm(incoming_id_to_anno):
        if args.trim_to_images_bboxes_labeled and annotations[0]['category_id'] == 5:
            # category_id 5 is No Object Visible
            continue

        anno_file_name = incoming.image_id_to_image[image_id]['file_name']
        dataset_name, seq_id, frame_num = file_name_to_parts(anno_file_name)

        im_rel_path = get_image_rel_path(dataset_seq_images, dataset_name, seq_id, frame_num)
        if im_rel_path is None:
            print(f'Not found in megadb entries: dataset {dataset_name},'
                  f' seq_id {seq_id}, frame_num {frame_num}')
            continue

        im_full_path = megadb_utils.get_full_path(datasets_table, dataset_name, im_rel_path)

        # download the image
        container_client = megadb_utils.get_storage_client(datasets_table, dataset_name)
        downloader = container_client.download_blob(im_full_path)
        image_file = io.BytesIO()
        blob_props = downloader.download_to_stream(image_file)
        image = vis_utils.open_image(image_file)

        boxes = [anno['bbox'] for anno in annotations]
        classes = [anno['category_id'] for anno in annotations]

        vis_utils.render_iMerit_boxes(boxes, classes, image, label_map=incoming.cat_id_to_name)

        file_name = '{}_gtbbox.jpg'.format(os.path.splitext(anno_file_name)[0].replace('/', '~'))
        image = vis_utils.resize_image(image, args.output_image_width)
        image.save(os.path.join(args.output_dir, 'rendered_images', file_name))

        images_html.append({
            'filename': '{}/{}'.format('rendered_images', file_name),
            'title': '{}, number of boxes: {}'.format(anno_file_name, len([b for b in boxes if len(b) > 0])),
            'textStyle': 
                'font-family:verdana,arial,calibri;font-size:80%;text-align:left;margin-top:20;margin-bottom:5'
        })

    # Write to HTML
    images_html = sorted(images_html, key=lambda x: x['filename'])
    write_html_image_list(
        filename=os.path.join(args.output_dir, 'index.html'),
        images=images_html,
        options={
            'headerHtml': '<h1>Sample annotations from {}</h1>'.format(args.incoming_annotation)
        })

    print('Visualized {} images.'.format(len(images_html)))


def main():
    parser = argparse.ArgumentParser()
    parser.add_argument(
        'incoming_annotation', type=str,
        help='Path to a json in the COCO format with relative coordinates for the bbox from annotators')
    parser.add_argument(
        'megadb_entries', type=str,
        help='Path to a json list of MegaDB entries to look up image path in blob storage')
    parser.add_argument(
        'output_dir', action='store', type=str,
        help='Output directory for html and rendered images')
    parser.add_argument(
        '--trim_to_images_bboxes_labeled', action='store_true',
        help='Only include images that have been sent for bbox labeling (but '
             'may be actually empty). Turn this on if QAing annotations.')
    parser.add_argument(
        '--num_to_visualize', action='store', type=int, default=200,
        help='Number of images to visualize. If trim_to_images_bboxes_labeled, there may be fewer than specified')
    parser.add_argument(
        '-w', '--output_image_width', type=int, default=1200,
        help='an integer indicating the desired width in pixels of the output '
             'annotated images. Use -1 to not resize.')

    args = parser.parse_args()

    assert 'COSMOS_ENDPOINT' in os.environ and 'COSMOS_KEY' in os.environ

    os.makedirs(args.output_dir, exist_ok=True)
    os.makedirs(os.path.join(args.output_dir, 'rendered_images'), exist_ok=True)

    visualize_incoming_annotations(args)


if __name__ == '__main__':
    main()



########
#
# load_api_results.py
#
# Loads the output of the batch processing API (json) into a pandas dataframe.
#
# Also functions to group entries by seq_id.
#
# Includes the deprecated functions that worked with the old CSV API output format.
#
########

#%% Constants and imports

from collections import defaultdict

import json
import os

from typing import Dict, Mapping, Optional, Tuple

import pandas as pd

from md_utils import ct_utils

headers = ['image_path', 'max_confidence', 'detections']


#%% Functions for grouping by sequence_id

def ss_file_to_file_name(f):
    # example
    # input 'file': 'SER/S1/F08/F08_R3/S1_F08_R3_PICT1150.JPG'
    # output 'id': 'S1/F08/F08_R3/S1_F08_R3_PICT1150.JPG'
    return f.split('SER/')[1].split('.JPG')[0]


def caltech_file_to_file_name(f):
    return f.split('cct_images/')[1].split('.')[0]


def api_results_groupby(api_output_path, gt_db_indexed, file_to_image_id, field='seq_id'):
    """
    Given the output file of the API, groupby (currently only seq_id).

    Args:
        api_output_path: path to the API output json file
        gt_db_indexed: an instance of IndexedJsonDb so we know the seq_id to image_id mapping
        file_to_image_id: a function that takes in the 'file' field in 'images' in the detector
            output file and converts it to the 'id' field in the gt DB.
        field: which field in the 'images' array to group by

    Returns:
    A dict where the keys are of the field requested, each points to an array
    containing entries in the 'images' section of the output file
    """

    with open(api_output_path) as f:
        detection_results = json.load(f)

    res = defaultdict(list)
    for i in detection_results['images']:
        image_id = file_to_image_id(i['file'])
        field_val = gt_db_indexed.image_id_to_image[image_id][field]
        res[field_val].append(i)
    return res


#%% Functions for loading the result as a Pandas DataFrame

def load_api_results(api_output_path: str, normalize_paths: bool = True,
                     filename_replacements: Optional[Mapping[str, str]] = None
                     ) -> Tuple[pd.DataFrame, Dict]:
    """
    Loads the json formatted results from the batch processing API to a
    Pandas DataFrame, mainly useful for various postprocessing functions.

    Args:
        api_output_path: path to the API output json file
        normalize_paths: whether to apply os.path.normpath to the 'file' field
            in each image entry in the output file
        filename_replacements: replace some path tokens to match local paths to
            the original blob structure

    Returns:
        detection_results: pd.DataFrame, contains at least the columns:
                ['file', 'detections','failure']            
        other_fields: a dict containing fields in the results other than 'images'
    """
    
    print('Loading API results from {}'.format(api_output_path))

    with open(api_output_path) as f:
        detection_results = json.load(f)

    # Validate that this is really a detector output file
    for s in ['info', 'detection_categories', 'images']:
        assert s in detection_results, 'Missing field {} in detection results'.format(s)

    # Fields in the API output json other than 'images'
    other_fields = {}
    for k, v in detection_results.items():
        if k != 'images':
            other_fields[k] = v

    # Normalize paths to simplify comparisons later
    if normalize_paths:
        for image in detection_results['images']:
            image['file'] = os.path.normpath(image['file'])
            # image['file'] = image['file'].replace('\\','/')

    # Replace some path tokens to match local paths to original blob structure
    if filename_replacements is not None:
        for string_to_replace in filename_replacements.keys():
            replacement_string = filename_replacements[string_to_replace]
            for im in detection_results['images']:
                im['file'] = im['file'].replace(string_to_replace,replacement_string)

    print('Converting results to dataframe')
    
    # If this is a newer file that doesn't include maximum detection confidence values,
    # add them, because our unofficial internal dataframe format includes this.
    for im in detection_results['images']:
        if 'max_detection_conf' not in im:
            im['max_detection_conf'] = ct_utils.get_max_conf(im)
    
    # Pack the json output into a Pandas DataFrame
    detection_results = pd.DataFrame(detection_results['images'])
    
        

    print('Finished loading API results for {} images from {}'.format(
            len(detection_results),api_output_path))

    return detection_results, other_fields


def write_api_results(detection_results_table, other_fields, out_path):
    """
    Writes a Pandas DataFrame back to a json that is compatible with the API output format.
    """

    print('Writing detection results to {}'.format(out_path))

    fields = other_fields

    images = detection_results_table.to_json(orient='records',
                                             double_precision=3)
    images = json.loads(images)
    fields['images'] = images

    with open(out_path, 'w') as f:
        json.dump(fields, f, indent=1)

    print('Finished writing detection results to {}'.format(out_path))


def load_api_results_csv(filename, normalize_paths=True, filename_replacements={}, nrows=None):
    """
    DEPRECATED
    Loads .csv-formatted results from the batch processing API to a pandas table
    """

    print('Loading API results from {}'.format(filename))

    detection_results = pd.read_csv(filename,nrows=nrows)

    print('De-serializing API results from {}'.format(filename))

    # Sanity-check that this is really a detector output file
    for s in ['image_path','max_confidence','detections']:
        assert s in detection_results.columns

    # Normalize paths to simplify comparisons later
    if normalize_paths:
        detection_results['image_path'] = detection_results['image_path'].apply(os.path.normpath)

    # De-serialize detections
    detection_results['detections'] = detection_results['detections'].apply(json.loads)

    # Optionally replace some path tokens to match local paths to the original blob structure
    # string_to_replace = list(options.detector_output_filename_replacements.keys())[0]
    for string_to_replace in filename_replacements:

        replacement_string = filename_replacements[string_to_replace]

        # iRow = 0
        for iRow in range(0,len(detection_results)):
            row = detection_results.iloc[iRow]
            fn = row['image_path']
            fn = fn.replace(string_to_replace,replacement_string)
            detection_results.at[iRow,'image_path'] = fn

    print('Finished loading and de-serializing API results for {} images from {}'.format(
        len(detection_results),filename))

    return detection_results


def write_api_results_csv(detection_results, filename):
    """
    DEPRECATED
    Writes a pandas table to csv in a way that's compatible with the .csv API output
    format.  Currently just a wrapper around to_csv that just forces output writing
    to go through a common code path.
    """

    print('Writing detection results to {}'.format(filename))

    detection_results.to_csv(filename, index=False)

    print('Finished writing detection results to {}'.format(filename))
